# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Records service component base classes."""

from invenio_records_resources.services.records.components import MetadataComponent


class DraftMetadataComponent(MetadataComponent):
    """Service component for draft metadata integration."""

    def update_draft(self, identity, data=None, record=None, **kwargs):
        """Update draft metadata."""
        record.metadata = data.get("metadata", {})

    def publish(self, identity, draft=None, record=None, **kwargs):
        """Update draft metadata."""
        record.metadata = draft.get("metadata", {})

    def edit(self, identity, draft=None, record=None, **kwargs):
        """Update draft metadata."""
        draft.metadata = record.get("metadata", {})

    def new_version(self, identity, draft=None, record=None, **kwargs):
        """Update draft metadata."""
        draft.metadata = record.get("metadata", {})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2023 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""UOW tests."""

from unittest.mock import MagicMock

import pytest
from invenio_records_resources.services.uow import UnitOfWork

from invenio_drafts_resources.services.records.uow import ParentRecordCommitOp

from .utils import create_and_publish


def test_parent_record_uow_params(app, db, service):
    """Test constructors params."""
    # test that it does not raise when params are passed correctly
    assert ParentRecordCommitOp({}, dict(service=service))
    assert ParentRecordCommitOp(
        {},
        indexer_context=dict(
            record_cls=service.record_cls,
            draft_cls=service.draft_cls,
            indexer=MagicMock(),
            draft_indexer=MagicMock(),
        ),
    )

    # test that it raises when one of the expected attr of the service is missing
    for missing_attr in ["record_cls", "draft_cls", "indexer", "draft_indexer"]:
        _service = type(
            "service",
            (object,),
            dict(
                record_cls=service.record_cls,
                draft_cls=service.draft_cls,
                indexer=service.indexer,
                draft_indexer=service.draft_indexer,
            ),
        )
        with pytest.raises(AttributeError):
            delattr(_service, missing_attr)
            assert ParentRecordCommitOp({}, indexer_context=dict(service=_service))

    # test that it raises when one of the expected constructor params is missing
    for missing_param in [
        "record_cls",
        "draft_cls",
        "indexer",
        "draft_indexer",
    ]:
        params = dict(
            record_cls=service.record_cls,
            draft_cls=service.draft_cls,
            indexer=service.indexer,
            draft_indexer=service.draft_indexer,
        )
        with pytest.raises(KeyError):
            params.pop(missing_param)
            assert ParentRecordCommitOp({}, indexer_context=dict(**params))


def test_parent_record_commit(app, db, service, identity_simple, input_data):
    """Test ParentRecordCommit unit of work."""
    record_v1 = create_and_publish(service, identity_simple, input_data)
    recid = record_v1.id
    draft_v1 = service.edit(identity_simple, recid)
    draft_v2 = service.new_version(identity_simple, recid)
    record_v2 = service.publish(identity_simple, draft_v2.id)

    parent = record_v1._record.parent

    indexer = MagicMock()
    draft_indexer = MagicMock()

    with UnitOfWork(db.session) as uow:
        op = ParentRecordCommitOp(
            parent,
            indexer_context=dict(
                record_cls=service.record_cls,
                draft_cls=service.draft_cls,
                indexer=indexer,
                draft_indexer=draft_indexer,
            ),
        )
        uow.register(op)
        uow.commit()

    assert indexer.bulk_index.call_count == 1
    args, _ = indexer.bulk_index.call_args
    assert list(args[0]) == [record_v1._record.id, record_v2._record.id]

    assert draft_indexer.bulk_index.call_count == 1
    args, _ = draft_indexer.bulk_index.call_args
    assert list(args[0]) == [draft_v1._record.id]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#

"""Base class for service components."""
from invenio_i18n import gettext as _
from invenio_records_resources.services.files.transfer import TransferType
from invenio_records_resources.services.records.components import (
    BaseRecordFilesComponent as _BaseRecordFilesComponent,
)
from invenio_records_resources.services.records.components import (
    ServiceComponent as BaseServiceComponent,
)
from marshmallow import ValidationError


class ServiceComponent(BaseServiceComponent):
    """Base service component."""

    def read_draft(self, identity, draft=None):
        """Update draft handler."""
        pass

    def update_draft(self, identity, data=None, record=None, errors=None):
        """Update draft handler."""
        pass

    def delete_draft(self, identity, draft=None, record=None, force=False):
        """Delete draft handler."""
        pass

    def edit(self, identity, draft=None, record=None):
        """Edit a record handler."""
        pass

    def new_version(self, identity, draft=None, record=None):
        """New version handler."""
        pass

    def publish(self, identity, draft=None, record=None):
        """Publish handler."""
        pass

    def import_files(self, identity, draft=None, record=None):
        """Import files handler."""
        pass

    def post_publish(self, identity, record=None, is_published=False):
        """Post publish handler."""
        pass


class BaseRecordFilesComponent(ServiceComponent, _BaseRecordFilesComponent):
    """Base record/draft files component.

    Attention: needs file attribute configuration!
    """

    #
    # API
    #
    def create(self, identity, data=None, record=None, errors=None):
        """Assigns files.enabled.

        NOTE: `record` actually refers to the draft
              (this interface is used in records-resources and rdm-records)
        """
        draft = record
        files = self.get_record_files(draft)
        enabled = data.get(self.files_data_key, {}).get(
            "enabled", self.service.config.default_files_enabled
        )

        if files.enabled != enabled:
            if not self.service.check_permission(
                identity, "manage_files", record=draft
            ):
                errors.append(
                    {
                        "field": f"{self.files_data_key}.enabled",
                        "messages": [
                            _("You don't have permissions to manage files options.")
                        ],
                    }
                )
                return  # exit early

        files.enabled = enabled

    def update_draft(self, identity, data=None, record=None, errors=None):
        """Assigns files.enabled and warns if files are missing.

        NOTE: `record` actually refers to the draft
              (this interface is used in records-resources and rdm-records)
        """
        draft = record
        draft_files = self.get_record_files(draft)
        default_preview = data.get(self.files_data_key, {}).get("default_preview")
        can_toggle_files = self.service.check_permission(
            identity, "manage_files", record=draft
        )

        enabled = data.get(self.files_data_key, {}).get(
            "enabled", self.service.config.default_files_enabled
        )

        if draft_files.enabled != enabled:
            if not can_toggle_files:
                errors.append(
                    {
                        "field": f"{self.files_data_key}.enabled",
                        "messages": [
                            _("You don't have permissions to manage files options.")
                        ],
                    }
                )
                return  # exit early

        try:
            self.assign_files_enabled(draft, enabled)
        except ValidationError as e:
            errors.append(
                {"field": f"{self.files_data_key}.enabled", "messages": e.messages}
            )
            return  # exit early

        if draft_files.enabled and not draft_files.items():
            if can_toggle_files:
                my_message = _(
                    "Missing uploaded files. To disable files for this record please mark it as metadata-only."
                )
            else:
                my_message = _("Missing uploaded files.")
            errors.append(
                {
                    "field": f"{self.files_data_key}.enabled",
                    "messages": [my_message],
                }
            )

        try:
            self.assign_files_default_preview(
                draft,
                default_preview,
            )
        except ValidationError as e:
            errors.append(
                {
                    "field": f"{self.files_data_key}.default_preview",
                    "messages": e.messages,
                }
            )

    def edit(self, identity, draft=None, record=None):
        """Edit callback."""
        draft_files = self.get_record_files(draft)
        record_files = self.get_record_files(record)
        draft_bucket = self.get_record_bucket(draft)
        lock_files = self.service.config.lock_edit_published_files(record)

        if draft_bucket is None:
            # Happens, when a soft-deleted draft is un-deleted.
            draft[self.files_data_key] = {"enabled": True}
            # re-fetch the files attribute - above data getter sets the attribute
            draft_files = self.get_record_files(draft)
            draft_files.create_bucket()

        # we copy always file objects and tear them down when publish
        draft_files.copy(record_files)
        if lock_files:
            # force going through the new version
            draft_files.lock()

    def new_version(self, identity, draft=None, record=None):
        """New version callback."""
        # We don't copy files from the previous version, but instead allow
        # users to import the files.
        draft_files = self.get_record_files(draft)
        record_files = self.get_record_files(record)
        draft_files.enabled = record_files.enabled

    def _purge_bucket_and_ovs(self, files):
        """Purge associated bucket and object versions."""
        if files.bucket:
            if files.bucket.locked:
                files.unlock()
            if files.enabled:
                files.delete_all(softdelete_obj=False, remove_rf=True)
            files.remove_bucket(force=True)

    def _publish_new(self, identity, draft, record):
        """Action when publishing a new draft."""
        # For unpublished drafts (new and new version), we move the draft
        # bucket from the draft to the record (instead of creating a new, and
        # deleting one). For consistency, we keep a bucket for all records
        # independently of if they have files enabled or not.
        record_files = self.get_record_files(record)
        draft_files = self.get_record_files(draft)
        draft_bucket = self.get_record_bucket(draft)

        record_files.set_bucket(draft_bucket)
        record_files.copy(draft_files, copy_obj=False)
        # Lock the bucket
        record_files.lock()

        # Cleanup
        if draft_files.enabled:
            # Hard delete all draft file records but keep the object versions
            draft_files.delete_all(remove_obj=False, remove_rf=True)
        draft_files.unset_bucket()

    def _publish_edit(self, identity, draft, record):
        """Action when publishing an edit to an existing record."""
        record_files = self.get_record_files(record)
        draft_files = self.get_record_files(draft)

        # sync draft files with record
        record_files.unlock()
        record_files.sync(draft_files, delete_extras=True)
        record_files.lock()

        # Teardown the bucket and files created in edit().
        self._purge_bucket_and_ovs(draft_files)

    def _check_file_completed(self, file_record):
        """Check if file upload has completed."""
        # Check if RDMDraftFile file has OV assigned
        # - if not, the upload is ongoing or has failed (fail handled elsewhere)
        # prevents ambiguous errors when trying to publish a record with
        # ongoing upload (cannot get storage class of None if OV is not set)
        has_attached_object = file_record.file is not None
        if not has_attached_object:
            return False
        transfer = TransferType(file_record.file.storage_class)
        if transfer.is_completed:
            return True

    def publish(self, identity, draft=None, record=None):
        """Copy bucket and files to record."""
        draft_files = self.get_record_files(draft)
        record_bucket_id = self.get_record_bucket_id(record)

        if draft_files.enabled and draft_files.bucket:
            if not draft_files.items():
                raise ValidationError(
                    _(
                        "Missing uploaded files. To disable files for "
                        "this record please mark it as metadata-only."
                    ),
                    field_name=f"{self.files_data_key}.enabled",
                )
        if draft_files.enabled:
            for file_record in draft_files.values():
                if not self._check_file_completed(file_record):
                    raise ValidationError(
                        _(
                            "One or more files have not completed their transfer, please wait."
                        ),
                        field_name=self.files_data_key,
                    )
        if record_bucket_id:
            self._publish_edit(identity, draft, record)
        else:
            self._publish_new(identity, draft, record)

    def delete_draft(self, identity, draft=None, record=None, force=False):
        """Delete files associated with a draft.

        :param force: If force is True, it means that the draft is being force
            deleted instead of soft deleted (i.e. an unpublished draft).
        """
        draft_files = self.get_record_files(draft)

        # Teardown the bucket and files if any.
        self._purge_bucket_and_ovs(draft_files)

    def import_files(self, identity, draft=None, record=None):
        """Import files handler."""
        record_files = self.get_record_files(record)
        draft_files = self.get_record_files(draft)
        if not draft_files.enabled:
            raise ValidationError(
                _("Files support must be enabled."),
                field_name=f"{self.files_data_key}.enabled",
            )

        if draft_files.items():
            raise ValidationError(
                _("Please remove all files first."),
                field_name=f"{self.files_data_key}.enabled",
            )

        if not record_files.enabled and not record_files.bucket:
            raise ValidationError(
                _("The record has no files."),
                field_name=f"{self.files_data_key}.enabled",
            )

        # Copy over the files
        draft_files.copy(record_files)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio Drafts Resources module to create REST APIs."""

from flask import g
from flask_resources import (
    JSONSerializer,
    ResponseHandler,
    resource_requestctx,
    response_handler,
    route,
    with_content_negotiation,
)
from invenio_records_resources.resources import RecordResource as RecordResourceBase
from invenio_records_resources.resources.records.headers import etag_headers
from invenio_records_resources.resources.records.resource import (
    request_data,
    request_extra_args,
    request_headers,
    request_read_args,
    request_search_args,
    request_view_args,
)
from invenio_records_resources.resources.records.utils import search_preference

from .errors import RedirectException


class RecordResource(RecordResourceBase):
    """Draft-aware RecordResource."""

    def create_blueprint(self, **options):
        """Create the blueprint."""
        # We avoid passing url_prefix to the blueprint because we need to
        # install URLs under both /records and /user/records. Instead we
        # add the prefix manually to each route (which is anyway what Flask
        # does in the end)
        options["url_prefix"] = ""
        return super().create_blueprint(**options)

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes

        def p(route):
            """Prefix a route with the URL prefix."""
            return f"{self.config.url_prefix}{route}"

        def s(route):
            """Suffix a route with the URL prefix."""
            return f"{route}{self.config.url_prefix}"

        rules = [
            route("GET", p(routes["list"]), self.search),
            route("POST", p(routes["list"]), self.create),
            route("GET", p(routes["item"]), self.read),
            route("PUT", p(routes["item"]), self.update),
            route("DELETE", p(routes["item"]), self.delete),
            route("GET", p(routes["item-versions"]), self.search_versions),
            route("POST", p(routes["item-versions"]), self.new_version),
            route("GET", p(routes["item-latest"]), self.read_latest),
            route("GET", p(routes["item-draft"]), self.read_draft),
            route("POST", p(routes["item-draft"]), self.edit),
            route("PUT", p(routes["item-draft"]), self.update_draft),
            route("DELETE", p(routes["item-draft"]), self.delete_draft),
            route("POST", p(routes["item-publish"]), self.publish),
            route("GET", s(routes["user-prefix"]), self.search_user_records),
        ]

        if self.service.draft_files:
            rules.append(
                route(
                    "POST",
                    p(routes["item-files-import"]),
                    self.import_files,
                    apply_decorators=False,
                )
            )

        return rules

    @request_extra_args
    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def search_user_records(self):
        """Perform a search over the record's versions.

        GET /user/records
        """
        hits = self.service.search_drafts(
            identity=g.identity,
            params=resource_requestctx.args,
            search_preference=search_preference(),
            expand=resource_requestctx.args.get("expand", False),
        )
        return hits.to_dict(), 200

    @request_extra_args
    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def search_versions(self):
        """Perform a search over the record's versions.

        GET /records/:pid_value/versions
        """
        hits = self.service.search_versions(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            params=resource_requestctx.args,
            search_preference=search_preference(),
            expand=resource_requestctx.args.get("expand", False),
        )
        return hits.to_dict(), 200

    @request_extra_args
    @request_view_args
    @response_handler()
    def new_version(self):
        """Create a new version.

        POST /records/:pid_value/versions
        """
        item = self.service.new_version(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 201

    @request_extra_args
    @request_view_args
    @response_handler()
    def edit(self):
        """Edit a record.

        POST /records/:pid_value/draft
        """
        item = self.service.edit(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 201

    @request_extra_args
    @request_view_args
    @response_handler()
    def publish(self):
        """Publish the draft."""
        item = self.service.publish(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 202

    @request_view_args
    @with_content_negotiation(
        response_handlers={
            "application/json": ResponseHandler(JSONSerializer(), headers=etag_headers)
        },
        default_accept_mimetype="application/json",
    )
    @response_handler(many=True)
    def import_files(self):
        """Import files from previous record version."""
        files = self.service.import_files(
            g.identity,
            resource_requestctx.view_args["pid_value"],
        )
        return files.to_dict(), 201

    @request_extra_args
    @request_view_args
    def read_latest(self):
        """Redirect to latest record.

        GET /records/:pid_value/versions/latest
        """
        item = self.service.read_latest(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            expand=resource_requestctx.args.get("expand", False),
        )
        raise RedirectException(item["links"]["self"])

    @request_extra_args
    @request_read_args
    @request_view_args
    @response_handler()
    def read_draft(self):
        """Edit a draft.

        GET /records/:pid_value/draft
        """
        item = self.service.read_draft(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 200

    @request_extra_args
    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def update_draft(self):
        """Update a draft.

        PUT /records/:pid_value/draft
        """
        item = self.service.update_draft(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            resource_requestctx.data or {},
            revision_id=resource_requestctx.headers.get("if_match"),
            expand=resource_requestctx.args.get("expand", False),
        )
        return item.to_dict(), 200

    @request_headers
    @request_view_args
    def delete_draft(self):
        """Delete a draft.

        DELETE /records/:pid_value/draft
        """
        self.service.delete_draft(
            g.identity,
            resource_requestctx.view_args["pid_value"],
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return "", 204



"""Example service."""

from invenio_records_resources.services import ConditionalLink
from invenio_records_resources.services import (
    FileServiceConfig as BaseFileServiceConfig,
)
from invenio_records_resources.services import RecordLink

from invenio_drafts_resources.services import RecordServiceConfig
from invenio_drafts_resources.services.records.components import (
    DraftFilesComponent,
    DraftMediaFilesComponent,
)
from invenio_drafts_resources.services.records.config import is_draft, is_record

from .api import Draft, DraftMediaFiles, Record, RecordMediaFiles
from .permissions import PermissionPolicy
from .schemas import RecordSchema


class ServiceConfig(RecordServiceConfig):
    """Mock service configuration."""

    permission_policy_cls = PermissionPolicy
    record_cls = Record
    draft_cls = Draft

    schema = RecordSchema

    components = RecordServiceConfig.components + [
        DraftFilesComponent,
        DraftMediaFilesComponent,
    ]

    links_item = {
        "self": ConditionalLink(
            cond=is_record,
            if_=RecordLink("{+api}/mocks/{id}"),
            else_=RecordLink("{+api}/mocks/{id}/draft"),
        ),
        "self_html": ConditionalLink(
            cond=is_record,
            if_=RecordLink("{+ui}/mocks/{id}"),
            else_=RecordLink("{+ui}/uploads/{id}"),
        ),
        "latest": RecordLink("{+api}/mocks/{id}/versions/latest"),
        "latest_html": RecordLink("{+ui}/mocks/{id}/latest"),
        "draft": RecordLink("{+api}/mocks/{id}/draft", when=is_record),
        "record": RecordLink("{+api}/mocks/{id}", when=is_draft),
        "publish": RecordLink("{+api}/mocks/{id}/draft/actions/publish", when=is_draft),
        "versions": RecordLink("{+api}/mocks/{id}/versions"),
    }


class MediaFilesRecordServiceConfig(RecordServiceConfig):
    """Record with media files service config."""

    service_id = "mock-record-media-files-service"
    record_cls = RecordMediaFiles
    draft_cls = DraftMediaFiles

    components = [
        DraftMediaFilesComponent,
    ]


class FileServiceConfig(BaseFileServiceConfig):
    """File service configuration."""

    allow_upload = False
    permission_policy_cls = PermissionPolicy
    record_cls = Record


class DraftFileServiceConfig(BaseFileServiceConfig):
    """File service configuration."""

    permission_policy_cls = PermissionPolicy
    permission_action_prefix = "draft_"
    record_cls = Draft


class MediaFileServiceConfig(BaseFileServiceConfig):
    """File service configuration."""

    service_id = "record-media-files-service"
    allow_upload = False
    permission_policy_cls = PermissionPolicy
    permission_action_prefix = "draft_media_"
    record_cls = RecordMediaFiles


class DraftMediaFileServiceConfig(BaseFileServiceConfig):
    """File service configuration."""

    service_id = "draft-media-files"
    permission_policy_cls = PermissionPolicy
    permission_action_prefix = "draft_media_"
    record_cls = DraftMediaFiles
